/**
  Lighting demo application.
  
  Orion Sky Lawlor, olawlor@acm.org, 2008-01-31 (Public Domain)
*/
#include <GL/glew.h> /*<- for gl...ARB extentions.  Must call glewInit after glutCreateWindow! */
#include <iostream>
#include <GL/glut.h>
#include <math.h>
#include "osl/vec4.h" /*<- handy little vector class */
/* Convert a struct with .x, .y, .z fields into three arguments */
#define VEC3_TO_XYZ(v) (v).x,(v).y,(v).z


#include "image.h" /* (originally from) Nigel Stewart's GLT image library */
#include "image.cpp"

#include "glsl.h" /* Orion's tiny GLSL library */
#include "glsl.cpp"

/*********************** App Code **********************/

/* Orthonormal coordinate frame */
class ortho_frame {
public:
	/* Unit vectors pointing along axes of our frame.
		X cross Y is Z. (right-handed coordinate system)
	*/
	vec3 x,y,z;
	ortho_frame() :x(1,0,0), y(0,1,0), z(0,0,1) {}
	
	/* Reorient this coordinate frame by this far 
	   along in the X and Y axes.  "dx" is the distance
	   along the z axis to push the x axis; "dy" 
	   the distance along the z axis to push the y axis.
	*/
	void nudge(double dx,double dy) {
		x+=dx*z;
		y+=dy*z;
		orthonormalize();
	}
	
	/* Reconstruct orthonormal frame from X and Y axes.
	  Y is primary, X is secondary, Z is tertiary.
	*/
	void orthonormalize(void) {
		y=normalize(y);
		z=normalize(cross(x,y));
		x=normalize(cross(y,z));
	}
};

/* Orientation of our camera. Camera looks down -Z axis; Y is up. */
ortho_frame camera_frame;

void my_display(void) {
	glMatrixMode(GL_PROJECTION); /* set up projection matrix */
	glLoadIdentity(); /* reset matrix at start of frame */
	gluPerspective(90.0, /* <- Y camera field-of-view, in degrees */
		glutGet(GLUT_WINDOW_WIDTH)/(float)glutGet(GLUT_WINDOW_HEIGHT), /* aspect ratio: width over height */
		0.01, /* Near clipping plane depth */
		100.0 /* Far clipping plane depth */
	);
	vec3 camera_location=1.3*camera_frame.z; /* back away from origin */
	gluLookAt(VEC3_TO_XYZ(camera_location),
		VEC3_TO_XYZ(vec3(0)),
		VEC3_TO_XYZ(camera_frame.y));
	
	glMatrixMode(GL_MODELVIEW); /* set up model and view matrix */
	glLoadIdentity(); /* reset matrix at start of frame */
	
	glClearColor(0.3,0.5,0.7,0); /* set background color to blue */
	glClear(GL_COLOR_BUFFER_BIT + GL_DEPTH_BUFFER_BIT);
	glEnable(GL_DEPTH_TEST);
	
	
	/* Tiny texturing demo */
	glActiveTexture(GL_TEXTURE0); /* talk to texture unit 0 */
	static int initted=0;
	if (0==initted++) /* only read texture once */
		readTextureFromFile("monkey.bmp");
	
	/* Enable fixed-function textures */
	glEnable(GL_TEXTURE_2D); 
	glColor4f(1,1,1,1);
	glBegin (GL_TRIANGLE_FAN);
	float z=0.2;
        glTexCoord2f(0,0); glVertex3d(0,0,z); 
        glTexCoord2f(1,0); glVertex3d(1,0,z); 
        glTexCoord2f(1,1); glVertex3d(1,1,z); 
        glTexCoord2f(0,1); glVertex3d(0,1,z); 
        glEnd();
	
	static GLhandleARB prog=makeProgramObject(
	"// GLSL Vertex shader\n"
	"varying vec4 reflectance;\n"
	"varying vec3 worldCoords;\n"
	"varying vec3 worldNormal;\n"
	"\n"
	"void main(void) {\n"
	"	reflectance = gl_Color;\n"
	"	worldCoords = vec3( gl_ModelViewMatrix * gl_Vertex );\n"
	"	worldNormal = normalize( gl_NormalMatrix * gl_Normal );\n"
	"	gl_Position = gl_ProjectionMatrix * gl_ModelViewMatrix * gl_Vertex; \n"
	"}\n"
	,
	"// GLSL Fragment shader\n"
	"varying vec4 reflectance;\n"
	"varying vec3 worldCoords;\n"
	"varying vec3 worldNormal;\n"
	"uniform vec3 cameraLoc;\n"
	"uniform sampler2D myTex;\n"
	"\n"
	"void main(void) {\n"
	"	vec3 wc = step(fract(8.0*worldCoords),vec3(0.5))*2.0-1.0;\n"
	"	float checker = clamp(wc.x*wc.y*wc.z,0.0,1.0);\n"
	"\n"	
	"	vec3 N = normalize(worldNormal); // Points away from surface\n"
	"	vec3 L1 = normalize(vec3(0,1,0)); // Points toward light 1\n"
	"	vec3 C = normalize(vec3(cameraLoc)-worldCoords); // Points toward camera\n"
	"	vec3 H1 = normalize(C+L1); // Blinn's 'halfway vector' for light 1\n"
	"\n"	
	"	float a=0.2; // Ambient illumination\n"
	"	float d=clamp(dot(N,L1),0.0,1.0); // Diffuse light fraction\n"
	"	float s=pow(clamp(dot(N,H1),0.0,1.0),50.0); // Specular fraction\n"
	"\n"
	"	/* Look up the color in our texture */ \n"
	"	vec4 tex = texture2D(myTex, vec2(worldCoords.x,worldCoords.y));\n"
	"	gl_FragColor = (a+d)*tex + vec4(s);\n"
	"}\n"
	);
	glUseProgramObjectARB(prog);
	
	glUniform3fvARB( /* set the "cameraCoords" uniform variable */
		glGetUniformLocationARB(prog, "cameraLoc"),
		1,
		camera_location
	);
	
	glUniform1iARB(glGetUniformLocationARB(prog,"myTex"),0);
	
	glColor3f(0.5,0.5,0.5); /* gray */
	glutSolidSphere(0.4, 25,50);
	
	glColor3f(1,1,0); /* yellow */
	glutSolidTorus(0.15,0.75, 25,50);
	
	/* Per-axis spheres */
	for (int axis=0;axis<3;axis++) {
		glPushMatrix();
		vec3 scale=vec3(0.25); /* squish the sphere */
		scale[axis]=1.0;  /* except this axis */
		glScalef(VEC3_TO_XYZ(scale)); 
		glColor3fv(scale); /* red, green, blue */
		glutSolidSphere(1.0,40,40);
		glPopMatrix();
	}
	glUseProgramObjectARB(0); /* back to fixed-function */
	
	glutSwapBuffers(); /* copy back buffer to front */
	
/* Statistics */
	static int last_msec=0;
	int cur_msec=glutGet(GLUT_ELAPSED_TIME); /* milliseconds since glutInit */
	int msec_elapsed=cur_msec-last_msec;
	printf("%d msec/frame; %.2f fps\n",
		msec_elapsed,1.0e3/msec_elapsed);
	last_msec=cur_msec;
}

int mouse_x, mouse_y;
void my_mouse(int state,int buttons,int x,int y) {
	mouse_x=x; mouse_y=y;  /* just stash motion */
}
void my_motion(int x,int y) { 
	float dx=x-mouse_x, dy=y-mouse_y; /* relative motion */
	
	float speed=1.0/100; /* radians rotation per pixel */
	camera_frame.nudge(speed*dx,-speed*dy); /* OpenGL Y is negative of mouse Y! */
	
	mouse_x=x; mouse_y=y; 
	glutPostRedisplay();
}

int main(int argc,char *argv[]) {
	glutInitWindowSize(800,600);
	glutInit(&argc,argv);
	int mode=GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH;
        glutInitDisplayMode (mode);
	glutCreateWindow("GLSL Lighting Demo");
	glewInit();
	
	glutDisplayFunc(my_display);
	glutMotionFunc(my_motion);
	glutMouseFunc(my_mouse);
	
	glutMainLoop(); /* processes events */
	return 0;
}
